﻿using System;
using System.Collections;
using System.Collections.Generic;
using System.Linq;
using Infragistics.Olap;

namespace FlatDataWithDistinctCount.Aggregators
{
    public enum CountBehavior
    {
        /// <summary>
        /// Evaluates the value with top most number of appearance.
        /// </summary>
        MaxCountValue,

        /// <summary>
        /// Evaluates the value with bottom most number of appearance.
        /// </summary>
        MinCountValue,

        /// <summary>
        /// Evaluates the max number of appearance.
        /// </summary>
        MaxCount,

        /// <summary>
        /// Evaluates the min number of appearance.
        /// </summary>
        MinCount,

        /// <summary>
        /// Evaluates the distinct count of the values.
        /// </summary>
        DistinctCount
    }

    public class CountAggregator<TInput, TResult> : IAggregator<TResult, CountCache<TInput, TResult>, TInput>
    {        
        public CountAggregator()
        {
            this.Behavior = CountBehavior.DistinctCount;
        }

        public AggregatorType AggregatorType
        {
            get { return AggregatorType.Calculated; }
        }

        public string Identity
        {
            get;
            set;
        }

        public CountBehavior Behavior { get; set; }

        public IAggregationResult<TResult, CountCache<TInput, TResult>> Evaluate(
            IAggregationResult<TResult, CountCache<TInput, TResult>> oldResult,
            IAggregateable aggregateable,
            IEnumerable items)
        {
            var result = new CountResult<TInput, TResult>(this.Behavior);
            foreach (var item in items)
            {
                var value = (TInput)aggregateable.GetValue(item);
                result.Cache.AddCount(value);
            }

            return result;
        }

        public IAggregationResult Evaluate(
            IAggregationResult oldResult,
            IAggregateable aggregateable,
            IEnumerable items)
        {
            return this.Evaluate((oldResult as CountResult<TInput, TResult>), aggregateable, items);
        }

        public IAggregationResult<TResult, CountCache<TInput, TResult>> Evaluate(
            IAggregationResult<TResult, CountCache<TInput, TResult>> oldResult,
            TInput value)
        {
            throw new NotImplementedException();
        }

        public IAggregationResult Evaluate(IAggregationResult oldResult, object value)
        {
            throw new NotImplementedException();
        }
    }

    public class CountResult<TInput, TResult> : IAggregationResult<TResult, CountCache<TInput, TResult>>
    {
        private readonly CountCache<TInput, TResult> _cache;

        public CountResult()
        {
            this._cache = new CountCache<TInput, TResult>();
        }

        public CountResult(CountBehavior behavior) :
            this()
        {
            this._cache.Behavior = behavior;
        }

        public CountCache<TInput, TResult> Cache
        {
            get
            {
                return this._cache;
            }
        }

        public TResult Value
        {
            get
            {
                return this.Cache.Value;
            }

            set
            {
                throw new NotImplementedException();
            }
        }

        object IAggregationResult.Value
        {
            get
            {
                return this.Value;
            }
        }

        object IAggregationResult.Cache
        {
            get
            {
                return this.Cache;
            }
        }
    }

    public class CountCache<TInput, TResult>
    {
        private IDictionary<TInput, int> _itemsCache = new Dictionary<TInput, int>();

        public CountCache()
        {
            this.Behavior = CountBehavior.DistinctCount;
        }

        public TResult Value
        {
            get
            {
                switch (this.Behavior)
                {
                    case CountBehavior.MaxCountValue:
                        int maxValue = this._itemsCache.Max(kvp => kvp.Value);
                        return (TResult)(object)this._itemsCache.Where(kvp => kvp.Value == maxValue).First().Key;

                    case CountBehavior.MaxCount:
                        return (TResult)(object)this._itemsCache.Max(kvp => kvp.Value);

                    case CountBehavior.MinCountValue:
                        int minValue = this._itemsCache.Min(kvp => kvp.Value);
                        return (TResult)(object)this._itemsCache.Where(kvp => kvp.Value == minValue).First().Key;

                    case CountBehavior.MinCount:
                        return (TResult)(object)this._itemsCache.Min(kvp => kvp.Value);

                    default:
                        return (TResult)(object)this._itemsCache.Count;
                }

            }
        }

        public void AddCount(TInput item)
        {
            ValueType value = item as ValueType;
            if (value == null && item == null)
            {
                return;
            }

            if (!this._itemsCache.ContainsKey(item))
            {
                this._itemsCache.Add(item, 1);
            }
            else
            {
                this._itemsCache[item]++;
            }
        }

        public CountBehavior Behavior { get; set; }
    }

    public class TopCountAggregator : CountAggregator<string, int>
    {
        public TopCountAggregator()
        {
            this.Behavior = CountBehavior.MaxCount;
        }
    }

    public class TopCountOfStringAggregator : CountAggregator<string, string>
    {
        public TopCountOfStringAggregator()
        {
            this.Behavior = CountBehavior.MaxCountValue;
        }
    }

    public class BottomCountOfStringAggregator : CountAggregator<string, string>
    {
        public BottomCountOfStringAggregator()
        {
            this.Behavior = CountBehavior.MinCountValue;
        }
    }

    public class CountAggregators : CountAggregator<string, int>
    {
        public CountAggregators()
        {
            this.Behavior = CountBehavior.DistinctCount;
        }
    }
}
